import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.awt.datatransfer.*; 
import java.awt.dnd.*; 
import java.io.File; 
import java.io.IOException; 
import java.awt.Component; 
import java.util.List; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class gokuraku_merge extends PApplet {

// 極楽画像計測（Gokuraku_measure）用データファイル結合プログラム(Gokuraku_merge)
// Created by Tatsuya Shirai
// National Institute of Technology, Suzuka college
// Mechanical Department
//
// Ver. 1.0 : 2022.06.08
// Ver. 2.0 : テキストファイル名の代わりに画像ファイル名を記録する（パス無し）

// 参考URL（ドラッグ＆ドロップに関して
// https://qiita.com/rain-lotus/items/258fa3defd41846bfd50
// https://sites.google.com/site/gutugutu30/other/processingdedoragguandodoroppu

Data         data;
ImportFile   im_file;
ExportFile   ex_file;
ShortMessage shortMessage;

final int Scr_w = 800;
final int Scr_h = 360;

public void setup() {
  data         = new Data();
  im_file      = new ImportFile();
  ex_file      = new ExportFile();
  shortMessage = new ShortMessage();

  //サイズ変更を許可する
  surface.setResizable(true);

  drop_init();
}

public void draw(){
  int x  = 16;
  int y  = 40;
  int dx = 100;
  int dy = 40;

  // 画面消去
  surface.setSize(Scr_w, Scr_h);
  fill(72, 72, 72);
  rect(0, 0, Scr_w - 1, Scr_h - 1);

  fill(180, 180, 255);
  // タイトル表示
  textSize(28);
  text("Gokuraku merge (@tatsuva) Ver." + Version, x, y); y += dy;

  // メニュー表示
  textSize(24);
  fill(180, 180, 255);
  text("[f]",    x, y); text(": select export folder", x + dx, y);  y += dy;
  text("[c]",    x, y); text(": clear data",           x + dx, y);  y += dy;
  text("[e]",    x, y); text(": export merged data",   x + dx, y);  y += dy;
  text("[ESC]",  x, y); text(": quit",                 x + dx, y);  y += dy;
  y += 16;
  textSize(16);
  text("Folder: " + ex_file.export_folder, x, y);                 y += dy;
  if (data.num > 0) {
                   text("Imported " + data.num + " files", x, y); y += dy;
  } else           text("Imported no files", x, y);               y += dy;
  shortMessage.display_short_message(x, y);                       y += dy;
  
}

public void keyPressed()
{
  switch (keyCode) {
    case 'f' :
    case 'F' :
      // エクスポートするフォルダーの選択
      File fp;
      fp = new File(ex_file.export_folder);
      selectFolder("Select a folder to process:", "folderSelected", fp);
      break;
    case 'c' :
    case 'C' : 
      // 全データ消去
      data.allclear();
      break;
    case 'e' :
    case 'E' : 
      // 全データのファイルへの出力
      ex_file.export_alldata(data);
      break;
  }
}

// エクスポートするフォルダーの選択
public void folderSelected(File selection) {
  if (selection == null) return;  // キャンセル時
  ex_file.set_export_folder(selection.getAbsolutePath());
  shortMessage.set_message("User select folder = " + ex_file.export_folder);
}

//// ドラッグ＆ドロップ関係
// 上記URL参照
// ドロップされたファイルの処理
public void fileSelected(List<File> fs) {
  for(File f:fs){
    // ドラッグ＆ドロップされたリストの中にフォルダーが含まれていた場合は，その層までは自動的に読み込む
    if (f.isDirectory()) {
      for (File sf:f.listFiles()) {
        if (sf.isFile()) im_file.read(sf.getAbsolutePath(), data);
      }
    } else im_file.read(f.getAbsolutePath(), data);
  }
}

  
  
  
  



DropTarget dropTarget;
Component  component;

public void drop_init() {  
  component = (Component)this.surface.getNative();
  dropTarget = new DropTarget(component, new DropTargetListener() {
    public void dragEnter(DropTargetDragEvent dtde) {
    }  
    public void dragOver(DropTargetDragEvent dtde) {
    }  
    public void dropActionChanged(DropTargetDragEvent dtde) {
    }  
    public void dragExit(DropTargetEvent dte) {
    }  
    public void drop(DropTargetDropEvent dtde) {  
      dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);  
      Transferable trans = dtde.getTransferable();  
      List<File> fileNameList = null;  
      if (trans.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {  
        try {  
          fileNameList = (List<File>)  
            trans.getTransferData(DataFlavor.javaFileListFlavor);
        } 
        catch (UnsupportedFlavorException ex) {
        } 
        catch (IOException ex) {
        }
      }  
      if (fileNameList == null) return;  
      fileSelected(fileNameList);
    }
  }
  );
}
// 定数
final int     Max_measure      = 32;            // １ファイルに含まれる最大のデータ数（実際のデータファイルの計測数がここで指定された数よりも少なくても構わない）
final String  DataFilename     = "merged.txt";  // 結合されたデータファイルのファイル名
final boolean Scale_convert    = true;          // スケールを考慮に入れて座標値を変換するか？（元に戻すか）
final boolean Output_color     = false;         // 各測定点の色データも結合するか？（一般的には不要なデータですね）

final int     Delay_message    = 5000;          // ショートメッセージの表示時間（ミリ秒）

final String  Version          = "2.0";
//// データ関係のクラス
// 結合されたデータのバッファー
class Data {
  OneData _top, _now;    // リスト構造の先頭と現在の書き込み・読み込みポインター
  int     num;           // 読み込まれたファイル数
  // コンストラクター
  Data() {
    _top = new OneData();
    gotoTop();
    num  = 0;
  }
  // リストの先頭に戻る
  public void gotoTop() {
    _now = _top;
  }
  // リストの次の要素に移動する
  public OneData next() {
    if (_now.next == null) return null;
    _now = _now.next;
    return _now;
  }
  // リストを全て削除する
  // （再起呼び出しを使った方がコードはシンプルだが…）
  public void allclear() {
    if (_top.next == null) return;
    println("Clear all data");
    gotoTop();
    while (_now.next != null) {
      OneData next;

      next       = _now.next;
      _now.next  = null;
      _now       = next;
    }
    _top = new OneData();
    gotoTop();
    num  = 0;
  }
  // １ファイルのデータをメモリにストア（リストを追加）する
  public OneData store(OneData onedata) {
    _now.next = _now.add(onedata);
    _now      = _now.next;
    num++;
    return _now;
  }
}

// １ファイル分のデータ
class OneData {  // リスト構造
  String filename;
  int    org_width, org_height;
  float  scale;
  int    num;
  int [] px    = new int[Max_measure];
  int [] py    = new int[Max_measure];
  int [] col_R = new int[Max_measure];
  int [] col_G = new int[Max_measure];
  int [] col_B = new int[Max_measure];
  OneData next;
  
  // コンストラクター
  OneData() {
    filename   = "";
    org_width  = 0; 
    org_height = 0;
    scale      = 0.0f;
    next       = null;
  }
  // データを書き込んで次のリストを作成する
  public OneData add(OneData onedata) {
    int i;
    filename   = onedata.filename;
    org_width  = onedata.org_width;
    org_height = onedata.org_height;
    scale      = onedata.scale;
    num        = onedata.num;
    for (i = 0; i < num; i++) {
      px[i]    = onedata.px[i];
      py[i]    = onedata.py[i];
      col_R[i] = onedata.col_R[i];
      col_G[i] = onedata.col_G[i];
      col_B[i] = onedata.col_G[i];
    }
    next = new OneData();
    return next;
  }
}

//// ファイル関係のクラス
//// ファイル関係のクラス
// CSV（タブ区切りの整数）のテキストを整数値に切り分けるクラス
class CsvStrings {
  String text;
  String str;
  float  value;
  // コンストラクター
  CsvStrings(String text) {
    this.set(text);
  }
  public void set(String text) {
    this.text = text + "\t";
    value     =  PApplet.parseFloat("NaN");  // floatの形式ではないのでNaNになる
  }
  // 次の値を取得する：正常に終了したらtrueを返し，valueに値が設定される
  public boolean next() {
    if (text.indexOf("\t") == -1) return false;
    if (text.length() == 0)       return false;
    if (text.equals("\t"))        return false;
    str   = text.substring(0, text.indexOf("\t"));
    value = PApplet.parseFloat(str);
    text  = text.substring(text.indexOf("\t") + 1);
    return true;
  }
  public boolean is_float() {
    if (value != value) return false;  // NaN, ちょっとトリッキーな方法
    return true;
  }
}

// 指定されたデータファイルから１画像分のデータを取得してメモリーに保存するクラス
class ImportFile {
  // コンストラクター
  ImportFile() {
  }

  // ファイル形式のチェック
  public boolean validation(String fname) {
    File         fp;
    String[]     lines;  // データファイルから読み出した全テキストデータ

    // 読み込めるか？
    if ((lines = loadStrings(fname)) == null) return false;
    // 少なくとも２行以上か？
    if ((lines.length < 2)) return false;
    // １行目はscale, org_width, org_heightか？
    CsvStrings csv = new CsvStrings(lines[0]);
    csv.next();
    if (! csv.is_float() || csv.value <= 0.0f)     return false;  // scale (float)
    csv.next();
    if (! csv.is_float() || PApplet.parseInt(csv.value) <= 0)  return false;  // org_width (int)
    csv.next();
    if (! csv.is_float() || PApplet.parseInt(csv.value) <= 0)  return false;  // org_height (int)
    // 互換性（Ver. 1.*のGokuraku_measureは画像ファイル名を記録していない）ので，この後に文字列があるかどうかはチェックしない
    // ２行目は0で始まるか？
    csv.set(lines[1]);
    csv.next();
    if (! csv.is_float() || PApplet.parseInt(csv.value) != 0)  return false;

    return true;
}
  
  // 実際にファイルのデータを読み込む
  public boolean read(String fname, Data data) {
    File         fp;
    CsvStrings   csv;
    OneData      onedata, now;
    String[]     lines;

    // ファイルチェック
    if (! validation(fname)) return false; 
    // ファイル読み込み
    if ((lines = loadStrings(fname)) == null) return false;

    // Scale, org_width, org_heightの読み出し
    onedata = new OneData();
    csv     = new CsvStrings(lines[0]);
    csv.next();
    onedata.scale      = csv.value;
    csv.next();
    onedata.org_width  = PApplet.parseInt(csv.value);
    csv.next();
    onedata.org_height = PApplet.parseInt(csv.value);
    csv.next();  // Ver.2 から，Gokuraku_maesureで画像ファイル名を記録するように変更
    if (csv.str.length() > 0) {
      onedata.filename      = csv.str;
    } else onedata.filename = "";

    // データ本体の読み込み
    for (int i = 1; i < lines.length; i++) {
      int value;
      csv.set(lines[i]);
      csv.next();
      if (PApplet.parseInt(csv.value) != (i - 1)) {
        shortMessage.set_message("Error: データ形式が正しくない（No.が一致しない）");
        return false;
      }
      // X座標
      csv.next();
      if (Scale_convert) onedata.px[i - 1] = PApplet.parseInt(csv.value / onedata.scale);
      else               onedata.px[i - 1] = PApplet.parseInt(csv.value);
      // Y座標
      csv.next();
      if (Scale_convert) onedata.py[i - 1] = PApplet.parseInt(csv.value / onedata.scale);
      else               onedata.py[i - 1] = PApplet.parseInt(csv.value);
      // R, G の空読み
      csv.next();
      onedata.col_R[i - 1] = PApplet.parseInt(csv.value);
      csv.next();
      onedata.col_G[i - 1] = PApplet.parseInt(csv.value);
      // B が存在するかどうかで一行のデータ数を確認
      if (csv.next() == false) {
        shortMessage.set_message("Error: データ形式が正しくない（一行のデータ数が少ない)");
        return false;
      }
      onedata.col_B[i - 1] = PApplet.parseInt(csv.value);
      onedata.num      = i;  // 実際に読み込んだ行数（更新する）
    }
    if (onedata.filename.length() == 0) {
      fp = new File(fname);  // ファイル名を分離するために．
      onedata.filename = fp.getName();
    }
    // データの格納
    now = data._now;
    data.store(onedata);
    shortMessage.set_message("Data imported: " + now.filename + " " + now.num + " points");
    if (now.next == null) {
      shortMessage.set_message("Error: データのメモリへの格納に失敗しました");
      return false;
    }
    return true;
  }     
}

// 結合したデータをファイルに出力するクラス
class ExportFile {
  String      export_folder;
  String      fullpathname;
  String      dest_filename;
  PrintWriter output;
  
  // コンストラクター
  ExportFile() {
    dest_filename = DataFilename;
    export_folder = sketchPath();
    set_export_folder(export_folder);
  }
  public void set_export_folder(String path) {
    if (path.equals("")) export_folder = sketchPath();  // この指定は無いはず
    else                 export_folder = path;
    File f = new File(export_folder, dest_filename);
    fullpathname = f.getAbsolutePath(); 
  }
  // 結合したデータのファイル出力
  public boolean export_alldata(Data data) {
    OneData now;
    if (data._top.next == null) return false;  // まだデータが無い

    output = createWriter(fullpathname);
    if (output == null) {
      shortMessage.set_message("Error: 出力ファイルが開けません : " + fullpathname);
      return false;
    } else println("Start exporting to " + fullpathname);

    // データファイルの出力
    int num = 0;;
    now = data._top;
    while (true) {
      // ヘッダーの書き込み
      output.print(now.filename);                                   // ファイル名
      output.print("\t" + now.scale);                               // 倍率（Scale_convertがtrueの場合はこの倍率でX, Y座標は変換されて結合される）
      output.print("\t" + now.org_width + "\t" + now.org_height);   // 画像の幅，高さ
      output.print("\t" + now.num);                                 // 計測点数
      // データ本体の書き込み
      for (int i = 0; i < now.num; i++) {
        output.print("\t" + now.px[i] + "\t" + now.py[i]);                                // X, Y座標
        // 色情報も出力する必要があるならば
        if (Output_color) {
          output.print("\t" + now.col_R[i] + "\t" + now.col_G[i] + "\t" + now.col_B[i]);  // R, G, B
        }
      }
      // 最後の改行
      output.println();
      num++;
      now = now.next;
      if (now.next == null) break;
    }    
    output.flush();
    output.close();
    shortMessage.set_message("Completde export " + num + " data files");
    data.allclear();
    return true;
  }
}

// ウインドウとコンソールエリアに短いメッセージを一時的に表示するためのクラス
class ShortMessage {
private  int    start;
private  String message;
  ShortMessage() {
    start   = millis();
    message = "";
  }
  public void set_message(String str) {
    message = str;
    println(str);
    start   = millis();
//  delay(500);          // 連続表示に対応するために．ただし動作が遅くなるため無効にしておくのでご自由に．
  }
  public void display_short_message(int x, int y) {
    if (millis() - start < Delay_message) {
      fill(255, 255, 255);
      textSize(16);
      text(message, x, y);
    }
  }
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "gokuraku_merge" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
